/* view history controller

Recently-visited views:

    The viewhistory object maintains a stack of recently visited lists. This
    information can be used to switch to recent views. The semantics of the
    following methods are as follows:
    - .getNextMostRecentView() and .getNextLeastRecentView() are used by the
      buffer switching commands to know which views to switch to.
    - .enterBufferSwitchingSession() and .exitBufferSwitchingSession() are
      called by the buffer swiching commands to help .setMostRecentView() know
      how to behave (the "recent views" stack is NOT updated during such a
      session). Note that .enter...() may be called multiple times for one
      .exit...() call.
    - .setMostRecentView() is called by the view manager's .selectView() and can
      be called directly by the buffer switching commands for proper session
      handling.

    NOTE: THIS BEHAVIOR IS TURNED OFF BY SETTING "handlestctrltab" to false on the tabbed view.
*/


function viewhistory() {
    this._recentViews = [];
    this.inBufferSwitchingSession = false;
    this._observingCtrlKeyUp = false;
    this.log = getLoggingMgr().getLogger('viewhistory');
    //this.log.setLevel(LOG_DEBUG);
    this._currentView = null;
    var obSvc = Components.classes["@mozilla.org/observer-service;1"].
                getService(Components.interfaces.nsIObserverService);
    obSvc.addObserver(this, 'view_closed',false);
    obSvc.addObserver(this, 'current_view_changed',false);
}

viewhistory.prototype.constructor = viewhistory;

viewhistory.prototype.doNextMostRecentView = function()
{
    this.log.info("doNextMostRecentView");
    var v = this.getNextMostRecentView();
    if (v != null) {
        this.enterBufferSwitchingSession();
        var obSvc = Components.classes["@mozilla.org/observer-service;1"].
                    getService(Components.interfaces.nsIObserverService);
        if (! this._observingCtrlKeyUp) {
            obSvc.addObserver(this, 'ctrl_keyup',false);
            this._observingCtrlKeyUp = true;
        }
        v.makeCurrent();
    }
}

viewhistory.prototype.doNextLeastRecentView = function()
{
    this.log.info("doNextLeastRecentView");
    var v = this.getNextLeastRecentView();
    if (v != null) {
        this.enterBufferSwitchingSession();
        var obSvc = Components.classes["@mozilla.org/observer-service;1"].
                    getService(Components.interfaces.nsIObserverService);
        if (! this._observingCtrlKeyUp) {
            obSvc.addObserver(this, 'ctrl_keyup',false);
            this._observingCtrlKeyUp = true;
        }
        v.makeCurrent();
    }
}

viewhistory.prototype._debug_recentViews = function()
{
    dump("Recent views: have " + this._recentViews.length + " views in history:\n");
    for (var i = 0; i < this._recentViews.length; i++) {
        dump('[' + i + ']: ');
        try {
            if (!this._recentViews[i]) {
                dump('<null>\n');
            } else {
                var v = this._recentViews[i].document.baseName;
                if (this._recentViews[i] == this._currentView) {
                    dump('*** ');
                }
                dump(v + '\n');
            }
        } catch (e) {
            this.log.error(e);
            dump("...\n");
        }
    }
}

viewhistory.prototype.getNextMostRecentView = function()
{
    if (! this._recentViews) { 
        log.error("no _recentViews in getNextMostRecentView");
        throw Components.results.NS_ERROR_FAILURE;
    }
    var currIndex = null;
    var i;
    for (i = 0; i < this._recentViews.length; i++) {
        if (this._recentViews[i] == this._currentView) {
            currIndex = i;
            break;
        }
    }
    // Find the current view in the "_recentViews" stack and return the
    // preceding view.
    var index = (currIndex + 1) % this._recentViews.length;
    var view = this._recentViews[index];
    if (typeof(view) == 'undefined') {
        // The "recent urls" list is out of date. Remove this entry and try
        // again.
        var newStack = new Array();
        for (i in this._recentViews) {
            i = Number(i);
            if (this._recentViews[i] != view) {
                newStack[newStack.length] = this._recentViews[i];
            }
        }
        this._recentViews = newStack;
        this.log.debug("getNextMostRecentView(): recents list is out of date, remove");
        return this.getNextMostRecentView();
    } else {
        return view;
    }
}

viewhistory.prototype.getNextLeastRecentView = function()
{
    if (! this._recentViews) { 
        log.error("no _recentViews in getNextLeastRecentView");
        throw Components.results.NS_ERROR_FAILURE;
    }
    // Find the current view in the "recent views" stack and return the
    // succeeding view.
    var currIndex = null;
    var i;
    for (i = 0; i < this._recentViews.length; i++) {
        if (this._recentViews[i] == this._currentView) {
            currIndex = i;
            break;
        }
    }
    // XXX JavaScript get's modulus wrong for negative values, so add the
    //     .length to ensure the value is not negative.
    var index = (currIndex - 1 + this._recentViews.length) %
                this._recentViews.length;

    var view = this._recentViews[index];
    if (view == undefined) {
        // The "recent urls" list is out of date. Remove this entry and try
        // again.
        var newStack = new Array();
        for (i in this._recentViews) {
            i = Number(i);
            if (this._recentViews[i] != view) {
                newStack[newStack.length] = this._recentViews[i];
            }
        }
        this._recentViews = newStack;
        //this.log.debug("getNextLeastRecentView(): curr=" + currUrl +
        //          ", recents list is out of date, remove " + url);
        return this.getNextLeastRecentView();
    } else {
        this.log.debug("TabViewManager.getNextLeastRecentView()");
        return view;
    }
}

viewhistory.prototype.enterBufferSwitchingSession = function()
{
    this.log.debug("enterBufferSwitchingSession()");
    this.inBufferSwitchingSession = true;
    gViewMgr.batchMode = true;
}

viewhistory.prototype.exitBufferSwitchingSession = function()
{
    this.log.debug("exitBufferSwitchingSession()");
    this.inBufferSwitchingSession = false;
    var observerSvc = Components.classes["@mozilla.org/observer-service;1"].
                       getService(Components.interfaces.nsIObserverService);
    gViewMgr.batchMode = false;
    // We need to do a current_view_changed here because the views.js
    // observer ignores those while in the middle of a control-tab session
    // for purposes of optimization.
    try {
        observerSvc.notifyObservers(this._currentView, 'current_view_changed', '');
    } catch (e) {
        // nobody home;
    }
}

viewhistory.prototype.observe = function (subject, topic, data)
{
    switch (topic) {
        case 'ctrl_keyup':
            var obSvc = Components.classes["@mozilla.org/observer-service;1"].
                        getService(Components.interfaces.nsIObserverService);
            obSvc.removeObserver(this, 'ctrl_keyup');
            this._observingCtrlKeyUp = false;
            this.exitBufferSwitchingSession();
            this.setMostRecentView(this._currentView);
            this._currentView.doFocus(); // needed to possibly start lint XXX want better API
            break;
        case 'view_closed':
            this.removeRecentView(subject);
            break;
        case 'current_view_changed':
            this._currentView = subject;
            break;
    }
}

viewhistory.prototype.setMostRecentView = function (view)
{
    //if (view) {
    //    this.log.debug("in setMostRecentView" + view.document.baseName);
    //}
    // Bring the given view to front of the "recent views" stack (note: it may
    // not yet be *in* the stack) *unless* we are currently in a buffer
    // switching session (e.g. press <Ctrl> and hold, hit <Tab> a few times,
    // release <Ctrl>).  Have to do the "unless" checking in here because
    // .selectView() calls this method.
    //this._debug_recentViews();
    var v;
    if (! this.inBufferSwitchingSession) {
        var newStack = new Array();
        newStack[0] = view;
        for (i =0; i < this._recentViews.length; i++) {
            v = this._recentViews[i];
            if (v != view) {
                newStack.push(v);
            }
        }
        this._recentViews = newStack;
        //this._debug_recentViews();
        //this.log.debug("setMostRecentView(" + view +
        //          "): new stack: " + this._recentViews);
    }
    else {
        this.log.debug("setMostRecentView(" + view + "): ignored because in buffer switching session");
    }
}


viewhistory.prototype.removeRecentView = function (view)
{
    // this._debug_recentViews();
    var newrecentviews = []
    
    for (i = 0; i < this._recentViews.length; i++) {
        if (this._recentViews[i] != view) {
           newrecentviews.push(this._recentViews[i]);
        }
    }
    this._recentViews = newrecentviews;    
    // this._debug_recentViews();
}

viewhistory.prototype.setState = function (state, viewcontainer)
{
    // This function sets a control-tab history from a pref
    // it is called by the onload handler that restores 'workspace'.
    // It stomps over whatever recent view history is established
    // by loading of views.  It needs to be called _after_
    // the views are available. It ignores URIs for views that
    // aren't loaded.
    this._recentViews = [];
    var marker = '?parentId=';
    var markerLength = marker.length;
    var identifier, URI, parentId, views, view;
    var index;
    for (i = 0; i < state.length; i++) {
        identifier = state.getStringPref(i);
        index = identifier.indexOf(marker);
        URI = identifier.slice(0, index);
        parentId = identifier.slice(index+markerLength);
        parent = document.getElementById(parentId);
        views = parent.findViewsForURI(URI);
        if (views.length > 0) {
            this._recentViews.push(views[0]);
        }
    }
}

viewhistory.prototype.getState = function ()
{
    // This function returns a pref which can be used in conjunction
    // with .setState() to serialize the 'recent views' history.
    var state = Components.classes['@activestate.com/koOrderedPreference;1'].createInstance();
    var index, v;
    for (var i = 0; i < this._recentViews.length; i++) {
        v = this._recentViews[i];
        if (v && v.parentView &&
            v.document && !v.document.isUntitled &&
            v.document.file && v.document.file.URI) {
            state.appendStringPref(v.document.file.URI + '?parentId=' + v.parentView.getAttribute('id'));
        }
    }
    return state;
}