/* Copyright (c) 2000-2003 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

// JavaScript-side support for Komodo's linting
//
// TODO:
// - document this module
// - currently if a filetype does not have an active linter, time is still
//   being wasted on lint manager calls, etc.


var gLint = {
    displayer     : null,
    service       : null,
    log           : null,
    globalPrefs   : null,
    observerSvc   : null
}

// jsLintBuffer implements koILintBuffer and koILintObserver
// and provides an abstraction for
// dealing with the virtual buffer/scimoz/py_inst mess.
function jsLintBuffer(view) {
    gLint.log.debug("create jsLintBuffer for view '"+view.title+
                    "' (lang '"+view.document.language+"')");
    this.view = view;
    this.errorString = null;

    // Register the lint buffer with the observer service to watch
    // for changes to the linting preference.
    gLint.service.observerService.addObserver(this, this.view.uid, false);

    this.lintResults = null;
    this._lintTimeout = null;
    this._clearTimeout = null;
    this._linterRequest = null;

    gLint.globalPrefs.prefObserverService.addObserver(this,'editUseLinting',false);
    this.view.prefs.prefObserverService.addObserver(this,'editUseLinting',false);
}

jsLintBuffer.prototype = {
    LINT_DELAY: 1000,  // milliseconds to wait before linting

    // nsIObserver implementation
    observe: function(subject, topic, data) {
        gLint.log.debug("observed "+subject+":"+topic+":"+data+" for view '"+
                        this.view.title+"' (uid="+this.view.uid+")");
        switch (topic) {
        case this.view.uid:
            var request = subject;
            if (data == 'lint_results') {
                this.handleResults(request);
            } else if (data == 'lint_clear') {
                this.handleClear(request);
            }
            break;
        case "editUseLinting":
            if (this.view.prefs.getBooleanPref("editUseLinting")) {
                this.request();
            }
            else {
                this.cancel();
                this.clear();
            }
        }
    },

    _notifyStart: function() {
        gLint.log.debug("jsLintBuffer._notifyStart()");
        try {
            gLint.observerSvc.notifyObservers(this.view,
                                              "current_view_check_started",
                                              "");
        } catch (ex) {
            // no observers are registered
        }
    },

    _lintRequest: function(linterType) {
        gLint.log.debug("jsLintBuffer._lintRequest(linterType='"+
                        linterType+"')");
        // Linters can't chdir to a remote URL, so don't pass in a working
        // directory unless it's local.
        var cwd = null;
        if (!this.view.document.isUntitled && this.view.document.file.isLocal) {
            cwd = this.view.document.file.dirName;
        }

        var lintRequest = Components.classes["@activestate.com/koLintRequest;1"].
                    createInstance(Components.interfaces.koILintRequest);
        lintRequest.document = this.view.document;
        lintRequest.linterType = linterType;
        lintRequest.uid = this.view.uid;
        lintRequest.cwd = cwd;
        lintRequest.results = null;
        lintRequest.errorString = '';
        lintRequest.canceled = false;
        
        return lintRequest;
    },
    
    _request: function() {
        gLint.log.debug("jsLintBuffer._request()");
        this._notifyStart();
        this._linterRequest = this._lintRequest(this.linterCID);
        gLint.service.addRequest(this._linterRequest);
        if (this._lintTimeout) {
            this._lintTimeout.stopTimeout();
            this._lintTimeout.free();
            this._lintTimeout = null;
        }
    },

    request: function() {
        gLint.log.debug("jsLintBuffer.request()");
        this.cancel();
        this.clear();
        if (!this._lintTimeout) {
            this._lintTimeout = new objectTimer(this,this._request,[]);
            this._lintTimeout.startTimeout(this.LINT_DELAY);
        }
    },

    _clear: function() {
        gLint.log.debug("jsLintBuffer._clear()");
        this.lintResults = null;
        if (this.linterCID != null) {
            gLint.service.broadcastResultsCleared(this._lintRequest(this.linterCID));
        }
    },

    clear: function() {
        gLint.log.debug("jsLintBuffer.clear()");
        // no need to clear if no results.
        if (this._linterRequest)
            this._linterRequest.canceled = true;
        if (this.lintResults) {
            if (!this._clearTimeout) {
                this._clearTimeout = new objectTimer(this,this._clear,[]);
                this._clearTimeout.startTimeout(0);
            }
        }
    },

    cancel: function() {
        gLint.log.debug("jsLintBuffer.cancel()");
        if (this._linterRequest)
            this._linterRequest.canceled = true;
        if (this._lintTimeout) {
            this._lintTimeout.stopTimeout();
            this._lintTimeout.free();
            this._lintTimeout = null;
        }
        if (this._clearTimeout) {
            this._clearTimeout.stopTimeout();
            this._clearTimeout.free();
            this._clearTimeout = null;
        }
        if (typeof(gLint.service) == 'undefined' || !gLint.service) {
            return;
        }
        gLint.service.cancelRequests(this.view.uid);
    },

    updateDisplay: function(startPos, styleLen) {
        gLint.log.debug("jsLintBuffer.updateDisplay(startPos="+startPos+
                        ", styleLen="+styleLen+")");
        if (this.lintResults) {
            gLint.displayer.display(this.view.scimoz, this.lintResults,
                                    this.view.document.languageObj.indicatorBits,
                                    startPos, styleLen);
        }
    },

    destructor: function() {
        gLint.log.debug("jsLintBuffer.destructor()");
        try {
            this.cancel();
        } catch(e) {
            gLint.log.exception(e);
        }
        try {
            // Stop observing changes in the editUseLinting preference.
            gLint.globalPrefs.prefObserverService.removeObserver(this,"editUseLinting");
            this.view.prefs.prefObserverService.removeObserver(this,"editUseLinting");
            if (gLint.service) {
                gLint.service.observerService.removeObserver(this,this.view.uid);
            }
        } catch(e) {
            gLint.log.debug(e);
        }
        this.view = null;
    },

    // private accessors
    // provide access to attributes that could be in the scimoz/mydata
    // pair or in a virtual buffer
    get language() {
        return this.view.document.language;
    },

    get linterCID() {
        var result = null;
        var linterLanguageService = this.view.document.languageObj.
            getLanguageService(Components.interfaces.koILinterLanguageService);
        if (linterLanguageService) {
            result = linterLanguageService.linterCID;
        }
        return result; 
    },

    // Methods called by the Lint Manager when interesting things happen
    // to our document.
    handleResults: function(lintRequest) {
        gLint.log.debug("jsLintBuffer.handleResults(lintRequest="+lintRequest+")");
        if (!lintRequest.canceled &&
            lintRequest.uid == this.view.uid) {
            this.lintResults = lintRequest.results;
            this.errorString = lintRequest.errorString;
            this.updateDisplay(0, this.view.scimoz.length)
            gLint.log.debug('notify current_view_check_finished');
            try {
                gLint.observerSvc.notifyObservers(this.view, "current_view_check_finished","");
            } catch(e) {
                // do nothing: Notify sometimes raises an exception if (???)
                // receivers are not registered?
            }
        }
    },

    handleClear: function(lintRequest) {
        gLint.log.debug("jsLintBuffer.handleClear(lintRequest="+lintRequest+")");
        if (lintRequest.uid == this.view.uid) {
            gLint.displayer.displayClear(this.view.scimoz,
                                         this.view.document.languageObj.indicatorBits);
        }
    },
    
    QueryInterface: function (iid) {
        if (!iid.equals(Components.interfaces.nsIObserver) &&
            !iid.equals(Components.interfaces.nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    }
    
}


function lint_initialize() {
    gLint.log = getLoggingMgr().getLogger("lint");
    //gLint.log.setLevel(LOG_DEBUG);
    gLint.log.debug("lint_initialize()");
    gLint.globalPrefs = Components.classes["@activestate.com/koPrefService;1"].
                    getService(Components.interfaces.koIPrefService).prefs;
    gLint.observerSvc = Components.classes["@mozilla.org/observer-service;1"].
                        getService(Components.interfaces.nsIObserverService);
    gLint.displayer = Components.classes["@activestate.com/koLintDisplayer;1"].
                    getService(Components.interfaces.koILintDisplayer);

    gLint.service = Components.classes["@activestate.com/koLintService;1"].
                  getService(Components.interfaces.koILintService);

    KomodoRegisterOnUnloadHandler(lint_unload);
}

function lint_unload() {
    gLint.log.debug("lint_unload()");
    if (gLint.service) {
        // free the lint manager
        gLint.service.terminate();
        gLint.service = null;
    }
}

//---- Jumping to lint results

function JumpToNextLintResult()
{
    try {
        gLint.log.debug("JumpToNextLintResult()");
        
        var view = gViewMgr.currentView;
        if (!view) return;
        if (!view.prefs.getBooleanPref("editUseLinting")) return;
        if (typeof(view.lintBuffer) == "undefined") return;
    
        // If there are no results, then say so.
        //   lintResults == null        -> inprogress
        //   lintResults.length == 0    -> ok
        //   lintResults.length > 0     -> errors/warnings
        var lintResults = view.lintBuffer.lintResults;
        if (view.lintBuffer.errorString) {
            dialog_alert(view.lintBuffer.errorString);
        } else if (!lintResults) {
            dialog_alert("Syntax checking is currently in progress.");
        } else if (! lintResults.getNumResults()) {
            dialog_alert("There are no syntax errors.");
        } else {
            // Determine the current position.
            var next = lintResults.getNextResult(view.currentLine,
                                                 view.currentColumn);
            //dump("next lint result: line="+next.lineStart+", column="+
            //     next.columnStart+"\n");
            
            var pos = view.scimoz.positionAtColumn(next.lineStart-1,
                                                   next.columnStart-1);
            view.scimoz.ensureVisibleEnforcePolicy(next.lineStart-1);
            view.scimoz.gotoPos(pos);
            view.scimoz.selectionStart = pos;
            view.scimoz.selectionEnd = pos;
        }
    } catch(ex) {
        gLint.log.exception(ex);
    }
}

