/* Copyright (c) 2000-2003 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

// Utility methods to easily use the KoIInterpolationService from JavaScript.
// Basically you can use the Interpolate_interpolate() method to iterpolation
// special '%'-escape codes in a given list of strings. See
// koInterpolationService.py (or play around with "run commands") for an
// authoritative description of available escape codes.
//
// How to interpolate a couple of strings (mystringa, mystringb). You can do as
// many strings as you want.
//
//    var imystringa = null;
//    var imystringb = null;
//    try {
//        var istrings = Interpolate_interpolate(
//                          <ref-to-komodo.xul>,
//                          [mystringa, mystringb],
//                          bracketed, // bool: codes are bracketed
//                          <optional-query-title>);
//        imystringa = istrings[0];
//        imystringb = istrings[1];
//    } catch (ex) {
//        var errno = lastErrorSvc.getLastErrorCode();
//        if (errno == Components.results.NS_ERROR_ABORT) {
//            // Command was cancelled.
//        } else if (errno == Components.results.NS_ERROR_INVALID_ARG) {
//            var errmsg = lastErrorSvc.getLastErrorMessage();
//            alert("Could not interpolate:" + errmsg);
//        } else {
//            log.error(ex);
//            alert("There was an unexpected error: " + ex);
//        }
//    }
//


//---- globals

var _interpolate_log = getLoggingMgr().getLogger("interpolate");

var _gISvc = Components.classes["@activestate.com/koInterpolationService;1"]
             .getService(Components.interfaces.koIInterpolationService);



//---- public interface

function Interpolate_isWordCharacter(ch)
{
    return /^\w/.test(ch);
}

function Interpolate_getWordUnderCursor(scimoz) {
    // Retrieve the current word under the cursor:
    var origCurrentPos = scimoz.currentPos;
    var origAnchor = scimoz.anchor;
    
    scimoz.anchor = scimoz.currentPos; // get rid of the selection
    if (scimoz.currentPos &&
        Interpolate_isWordCharacter(scimoz.getWCharAt(scimoz.currentPos-1))) {  // There is part of a word to our left
        scimoz.wordLeft();
    }
    // Using several wordPartRights instead of one wordRight because the latter is whitespace swallowing.
    var sentinel = scimoz.currentPos;
    while (scimoz.currentPos < scimoz.textLength &&
           Interpolate_isWordCharacter(scimoz.getWCharAt(scimoz.currentPos))) {  // There is part of a word to our right
        scimoz.wordPartRightExtend();
        if (sentinel == scimoz.currentPos) { // we didn't move!  break out to avoid infinite loop
            break;
        }
    }
    var word = scimoz.selText;
    scimoz.currentPos = origCurrentPos;
    scimoz.anchor = origAnchor;
    return word;
}

// A utility function to retrieving specific view data required by
// Interpolate_interpolate().
//  "editor" is a reference to the komodo.xul window.
//  "viewData" (optional) is a object possibly containing overriding values for
//      some of the view data.
function Interpolate_getViewData(editor,
                                 viewData /* =<determined from curr view> */)
{
    if (typeof viewData == 'undefined' || viewData == null)
        viewData = new Object();

    var view = editor.gViewMgr.currentView;
    var scimoz = null;
    if (view && view.getAttribute("type") == "editor") {
        scimoz = view.scintilla.scimoz;
    }

    // fileName
    if ( !("fileName" in viewData) ) {
        if (view) {
            viewData.fileName = view.document.displayPath;
        } else {
            viewData.fileName = null;
        }
    }

    // lineNum
    if ( !("lineNum" in viewData) ) {
        if (scimoz) {
            viewData.lineNum = scimoz.lineFromPosition(scimoz.currentPos) + 1;
        } else {
            viewData.lineNum = 0;
        }
    }

    // selection
    if ( !("selection" in viewData) ) {
        if (scimoz) {
            viewData.selection = scimoz.selText;
        } else {
            viewData.selection = null;
        }
    }

    // word
    if ( !("word" in viewData) ) {
        if (scimoz) {
            viewData.word = Interpolate_getWordUnderCursor(scimoz);
        } else {
            viewData.word = null;
        }
    }

    return viewData;
}


// Interpolate '%'-escape codes in the given list of strings.
//
//  "editor" is a reference the komodo.xul window.
//  "strings" is a list of raw strings to interpolate.
//  "bracketed" is a boolean indicating if the interpolation codes are
//      bracketed.
//  "queryTitle" (optional) is a title for the possible query dialog raised
//      during interpolation.
//  "viewData" (optional) allows one to override specific view data used for
//      interpolation. By default view data is retrieved from the current view.
//      This may not always be appropriate. It may be an object with one or
//      more of the following attributes:
//          "fileName" is the filename of the current file (null, if N/A);
//          "lineNum" is the current line number (0, if N/A);
//          "word" is the current word under cursor (null if none);
//          "selection" is the current selection (null if none).
//
//  Returns a list of interpolated strings on success, otherwise an exception
//  is raised and an error set on the last error service:
//      koILastError errno      reason
//      ----------------------- -----------------------------------------
//      NS_ERROR_ABORT          User cancelled the query dialog.
//      NS_ERROR_INVALID_ARG    A normal interpolation failure because of
//                              invalid interp code usage.
//
function Interpolate_interpolate(editor, strings, bracketed,
                                 queryTitle /* =null */,
                                 viewData /* =<determined from current view> */)
{
    if (typeof queryTitle == 'undefined') queryTitle = null;
    _interpolate_log.info("Interpolate_interpolate(editor, strings=["+strings+
                          "], bracketed="+bracketed+", queryTitle='"+
                          queryTitle+"', viewData)");
    viewData = Interpolate_getViewData(editor, viewData);

    var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                       .getService(Components.interfaces.koILastErrorService);

    // Interpolation step 1: get queries.
    var queriesCountObj = new Object();
    var queriesObj = new Array();
    var i1countObj = new Object();
    var i1stringsObj = new Array();
    _gISvc.Interpolate1(strings.length, strings,
                        viewData.fileName, viewData.lineNum,
                        viewData.word, viewData.selection,
                        bracketed,
                        queriesCountObj, queriesObj,
                        i1countObj, i1stringsObj);
    var queries = queriesObj.value;
    var istrings = i1stringsObj.value;

    // Ask the user for required data. (If there are no queries then we are
    // done interpolating.)
    if (queries.length != 0) {
        var obj = new Object();
        obj.queries = queries;
        obj.title = queryTitle;
        openWindowMultipleInstance("chrome://jaguar/content/run/interpolationquery.xul", 
                                   "komodo_interpolationquery",
                                   // 'all' allows resizing
                                   "chrome,modal",
                                   obj);
        if (obj.retval == "Cancel") {
            var errmsg = "Interpolation query cancelled.";
            lastErrorSvc.setLastError(Components.results.NS_ERROR_ABORT,
                                      errmsg);
            throw errmsg;
        }

        // Interpolation step 2: interpolated with answered queries.
        var i2countObj = new Object();
        var i2stringsObj = new Array();
        _gISvc.Interpolate2(istrings.length, istrings,
                            queries.length, queries,
                            i2countObj, i2stringsObj);
        istrings = i2stringsObj.value;
    }

    _interpolate_log.info("Interpolate_interpolate: istrings=["+istrings+"]");
    return istrings;
}


// Return a list of start and end offsets of interpolation code blocks into the
// given strings "s".
function Interpolate_getBlockOffsets(s, bracketed)
{
    var offsetsCountObj = new Object();
    var offsetsObj = new Array();
    _gISvc.GetBlockOffsets(s, bracketed, offsetsCountObj, offsetsObj);
    var offsets = offsetsObj.value;

    return offsets;
}


