#!/usr/bin/env python
# Copyright (c) 2000-2001 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

import sys, os, re, string
import tempfile
from xpcom import components, ServerException, COMException, nsError
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC
import process
import koprocessutils
import which
import logging

log = logging.getLogger('koAppInfo')

#---- components
class KoAppInfoEx:
    def __init__(self):
        self.installationPath = ''
        self.executablePath = ''
        self.haveLicense = 0
        self.buildNumber = 0
        self.localHelpFile = ''
        self.webHelpURL = ''
        self._configPath = ''
        self.installed = 0

        self._proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        self._prefSvc = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService)
        self.prefService = self._proxyMgr.getProxyForObject(None,
            components.interfaces.koIPrefService, self._prefSvc,
            PROXY_ALWAYS | PROXY_SYNC)

        
    def FindInstallationPaths(self):
        return []
    
    def getInstallationPathFromBinary(self, binaryPath):
        return ''
    
    def set_installationPath(self, path):
        self.installationPath = path
        self.executablePath = ''
    
    # pulled over from koIInterpreterLanguageService for BC
    def get_interpreterPath(self):
        return self.get_executablePath()
    
    # pulled over from koIInterpreterLanguageService for BC
    def get_includePath(self):
        return self._configPath

class KoPerlInfoEx(KoAppInfoEx):
    _com_interfaces_ = [components.interfaces.koIPerlInfoEx]
    _reg_clsid_ = "adb73505-eed5-46c5-8425-ce0bd8a5ec47"
    _reg_contractid_ = "@activestate.com/koPerlInfoEx;1"
    _reg_desc_ = "Extended Perl Information"
    
    def __init__(self):
        KoAppInfoEx.__init__(self)
        self._userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)

    def _GetPerlExeName(self):
        if not self.installationPath:
            perlExe = self.prefService.prefs.getStringPref("perlDefaultInterpreter")
            if perlExe: return perlExe
            self.installationPath = self.FindInstallationPaths()[0]
        assert self.installationPath is not None

        if sys.platform.startswith("win"):
            return os.path.join(self.installationPath, "bin", "perl.exe")
        else:
            return os.path.join(self.installationPath, "bin", "perl")

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        if sys.platform.startswith('win'):
            exts = ['.exe']
        else:
            exts = None
        perlExes = which.whichall("perl", exts=exts, path=self._userPath)
        perlInstallationPaths = [self.getInstallationPathFromBinary(p)\
                                 for p in perlExes]
        return perlInstallationPaths

    def getInstallationPathFromBinary(self, binaryPath):
        return os.path.dirname(os.path.dirname(binaryPath))

    def get_executablePath(self):
        return self._GetPerlExeName()

    def get_haveLicense(self):
        return 1

    def get_version(self):
        perlExe = self._GetPerlExeName()
        if not os.path.exists(perlExe):
            raise ServerException(nsError.NS_ERROR_FILE_NOT_FOUND)
        argv = [perlExe, "-v"]
        p = process.ProcessOpen(argv)
        perlVersionDump = p.stdout.read()
        p.close()
        # Old perls look like: This is perl, version 5.005_03 built for MSWin32-x86-object
        # New perls look like: This is perl, v5.6.1 built for MSWin32-x86-multi-thread
        perlVersionMatch = re.search("This is perl, v(?:ersion )?([0-9._]+)", perlVersionDump)
        if perlVersionMatch:
            return perlVersionMatch.group(1)
        return ''

    def get_buildNumber(self):
        argv = [self.get_executablePath(), "-v"]
        p = process.ProcessOpen(argv)
        versionDump = p.stdout.read()
        p.close()
        pattern = re.compile("Binary build (\d+) provided by ActiveState")
        match = pattern.search(versionDump)
        if match:
            return int(match.group(1))
        else:
            # This is likely not an ActivePerl installation.
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)
 
    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        If there is an html/index.html in the install directory found
        via `which perl`. An *Active*Perl installation could be found via the
        registry on Windows (see get_installed()) since this is the only type
        of Perl installation likely to have the html/index.html subfile.
        However the current suffices.
        """
        try:
            perlExe = which.which("perl", path=self._userPath)
        except which.WhichError:
            return None

        indexHtml = os.path.join(os.path.dirname(perlExe),
                                 "..", "html", "index.html")
        if os.path.isfile(indexHtml):
            return indexHtml
        else:
            return None

    def get_webHelpURL(self):
        """Return a web URL for help on this app, else return None."""
        return "http://aspn.activestate.com/ASPN/Products/ASPNTOC-ACTIVEPERL"
    
    # koIPerlInfoEx routines
    def haveModules(self, modules):
        perlExe = self.get_executablePath()
        for module in modules:
            argv = [perlExe, "-M"+module, "-e1"]
            p = process.ProcessOpen(argv)
            retval = p.wait()
            p.close()
            if retval: # if returns non-zero, then don't have that module
                return 0
        else:
            return 1

class KoPythonInfoEx(KoAppInfoEx):
    _com_interfaces_ = [components.interfaces.koIAppInfoEx]
    _reg_clsid_ = "b76bc2ee-261e-4597-b1ef-446e9bb89d7c"
    _reg_contractid_ = "@activestate.com/koPythonInfoEx;1"
    _reg_desc_ = "Extended Python Information"
    
    def __init__(self):
        KoAppInfoEx.__init__(self)
        self.koInfoService = components.classes["@activestate.com/koInfoService;1"].getService();
        self._userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)
        
    def _GetPythonExeName(self):
        if not self.installationPath:
            pythonExe = self.prefService.prefs.getStringPref("pythonDefaultInterpreter")
            if pythonExe: return pythonExe
            paths = self.FindInstallationPaths()
            if paths:
                self.installationPath = paths[0]
            else:
                return None
        assert self.installationPath is not None

        if sys.platform.startswith("win"):
            if self.koInfoService.buildType == "release":
                return os.path.join(self.installationPath, "python.exe")
            else:
                return os.path.join(self.installationPath, "python_d.exe")
        else:
           return os.path.join(self.installationPath, "bin", "python")

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        if sys.platform.startswith('win'):
            exts = ['.exe']
        else:
            exts = None
        pythonExes = which.whichall("python", exts=exts, path=self._userPath)
        pythonInstallationPaths = [self.getInstallationPathFromBinary(p)\
                                   for p in pythonExes]
        return pythonInstallationPaths

    def getInstallationPathFromBinary(self, binaryPath):
        if sys.platform.startswith("win"):
            return os.path.dirname(binaryPath)
        else:
            return os.path.dirname(os.path.dirname(binaryPath))

    def get_executablePath(self):
        return self._GetPythonExeName()

    def get_haveLicense(self):
        return 1

    def get_version(self):
        # python 1.52 does not support the -V option
        versionScript = \
        r"import sys;"\
        r"import string;"\
        r"paren = string.find(sys.version, '(');"\
        r"version = string.strip(sys.version[:paren]);"\
        r"sys.stdout.write(version)"

        pythonExe = self._GetPythonExeName()
        pythonDir = os.path.dirname(pythonExe)
        argv = [pythonExe, "-c", versionScript]
        p = process.ProcessOpen(argv, cwd=pythonDir)
        versionDump = p.stdout.read()
        p.close()
        # we are only to rely on the first 2 digits being in the form x.y
        match = re.search("([0-9.]+)", versionDump)
        if match:
            return match.group(1)
        return ''

    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        Windows:
            If there is a
            'HKLM/Software/Python/PythonCore/<major>.<minor>/Help/Main Python Documentation'
            and if the identified file exists.
        Linux/Solaris:
            If there is an html/index.html in the install directory found
            via `which python`.
        """
        if sys.platform.startswith("win"):
            import _winreg
            # get the base PythonCore registry key
            try:
                pythonCoreKey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                                                "SOFTWARE\\Python\\PythonCore")
            except EnvironmentError:
                return None
            # get a list of each installed version 
            versions = []
            index = 0
            while 1:
                try:
                    versions.append(_winreg.EnumKey(pythonCoreKey, index))
                except EnvironmentError:
                    break
                index += 1
            # try to find a existing help file (prefering the latest
            # installed version)
            versions.sort()
            versions.reverse()
            for version in versions:
                try:
                    helpFileKey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                        "SOFTWARE\\Python\\PythonCore\\%s\\Help\\Main "\
                        "Python Documentation" % version)
                    helpFile, keyType = _winreg.QueryValueEx(helpFileKey, "")
                    if os.path.isfile(helpFile):
                        return helpFile
                except EnvironmentError:
                    pass
            return None
        else:
            try:
                pythonExe = which.which("python", path=self._userPath)
            except which.WhichError:
                return None
            indexHtml = os.path.join(os.path.dirname(pythonExe),
                                     "..", "html", "index.html")
            if os.path.isfile(indexHtml):
                return indexHtml
            else:
                return None

    def get_webHelpURL(self):
        """Return a web URL for help on this app, else return None."""
        return "http://aspn.activestate.com/ASPN/Products/ASPNTOC-ACTIVEPYT_"


class KoTclInfoEx(KoAppInfoEx):
    _com_interfaces_ = [components.interfaces.koITclInfoEx]
    _reg_clsid_ = "DF64A66F-FD69-4F5E-92B2-B3C9F8638F66"
    _reg_contractid_ = "@activestate.com/koTclInfoEx;1"
    _reg_desc_ = "Extended Tcl Information"

    def __init__(self):
        KoAppInfoEx.__init__(self)
        self._userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)

    def get_executablePath(self):
        # XXX invoke interpreters has logic for using wish, do we need
        # it here also?
        if not self.installationPath:
            tclExe = self.prefService.prefs.\
                     getStringPref("tclshDefaultInterpreter")
            if tclExe: return tclExe
            self.installationPath = self.FindInstallationPaths()[0]
        assert self.installationPath is not None

        if sys.platform.startswith("win"):
            return os.path.join(self.installationPath, "tclsh.exe")
        else:
           return os.path.join(self.installationPath, "bin", "tclsh")

    def _getTclshExeName(self):
        if sys.platform.startswith('win'):
            return 'tclsh.exe'
        else:
            return 'tclsh'

    def _getWishExeName(self):
        if sys.platform.startswith('win'):
            return 'wish.exe'
        else:
            return 'wish'

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        if sys.platform.startswith('win'):
            exts = ['.exe']
        else:
            exts = None
        tclshs = which.whichall("tclsh", exts=exts, path=self._userPath)
        installPaths = [self.getInstallationPathFromBinary(tclsh)\
                        for tclsh in tclshs]
        uniqueInstallPaths = {}
        for installPath in installPaths:
            uniqueInstallPaths[installPath] = 1
        installPaths = uniqueInstallPaths.keys()
        installPaths.sort()
        return installPaths

    def _isInstallationLicensed(self, installationPath):
        return 1
    
    def get_haveLicense(self):
        return self._isInstallationLicensed(self.installationPath)

    def get_version(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
 
    def get_buildNumber(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
 
    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        Windows:
            If there is a 'HKLM\Software\ActiveState\ActiveTcl\<CurVer>\Help'
            and if the identified file exists.
        Linux/Solaris:
            Nada. Just man files, which I don't consider "launchable" in a
            browser context. XXX Perhaps they *are* in Nautilus? 
        """
        if sys.platform.startswith("win"):
            import _winreg
            # get the base ActiveTcl registry key
            try:
                activeTclKey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                                               "SOFTWARE\\ActiveState\\ActiveTcl")
            except EnvironmentError:
                return None
            # get a list of each installed version 
            versions = []
            index = 0
            while 1:
                try:
                    versions.append(_winreg.EnumKey(activeTclKey, index))
                except EnvironmentError:
                    break
                index += 1
            # try to find a existing help file (prefering the latest
            # installed version)
            versions.sort()
            versions.reverse()
            for version in versions:
                try:
                    helpFileKey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                        "SOFTWARE\\ActiveState\\ActiveTcl\\%s\\Help" % version)
                    helpFile, keyType = _winreg.QueryValueEx(helpFileKey, "")
                    if os.path.isfile(helpFile):
                        return helpFile
                except EnvironmentError:
                    pass
        return None

    def get_webHelpURL(self):
        return "http://aspn.activestate.com/ASPN/Products/ASPNTOC-ACTIVETCL_"

    def getInstallationPathFromBinary(self, binaryPath):
        return os.path.dirname(os.path.dirname(binaryPath))

    def selectDefault(self):
        paths = self.FindInstallationPaths()
        
        for installationPath in paths: 
            if self._isInstallationLicensed(installationPath):
                self.installationPath = installationPath
                return 1
        else:
            # Otherwise use whatever is left
            if paths:
                self.installationPath = paths[0]
                return 1
            else:
                self.installationPath = None
                return 0
            
    def get_tclsh_path(self):
        return os.path.join(self.installationPath, "bin",
                            self._getTclshExeName())
        
    def get_wish_path(self):
        return os.path.join(self.installationPath, "bin",
                            self._getWishExeName())

class KoPHPInfoInstance(KoAppInfoEx):
    _com_interfaces_ = [components.interfaces.koIPHPInfoEx]
    _reg_clsid_ = "E2066A3A-FC6D-4157-961E-E03C020594BE"
    _reg_contractid_ = "@activestate.com/koPHPInfoInstance;1"
    _reg_desc_ = "PHP Information"

    # the purpose of KoPHPInfoInstance is to be able to define
    # what executable and ini path are used without prefs getting
    # in the way.  If you want to use prefs, use koPHPInfoEx.
    def __init__(self):
        KoAppInfoEx.__init__(self)
        self._configPath = None
        self._executable = None
        self._userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)

    def _findPHP(self):
        if not self.installationPath:
            paths = self.FindInstallationPaths()
            if paths:
                self.installationPath = paths[0]
            else:
                return None
        
        for phpname in ['php','php4','php-cgi','php-cli']:
            if sys.platform.startswith("win"):
                phpname += '.exe'
            php = os.path.join(self.installationPath, phpname)
            if os.path.exists(php):
                break

        return php
        
    def _GetPHPExeName(self):
        if self._executable:
            return self._executable
        return self._findPHP()

    def _getInterpreterConfig(self):
        if self._configPath:
            return self._configPath
        return None

    def _GetPHPOutput(self, phpCode):
        """Run the given PHP code and return the output.

        If some error occurs then the error is logged and the empty
        string is returned. (Basically we are taking the position that
        PHP is unreliable.)
        """
        php = self._GetPHPExeName()
        if not php:
            return None #XXX Would be better, IMO, to raise an exception here.
        env = koprocessutils.getUserEnv()
        ini = self._getInterpreterConfig()
        if ini:
            env["PHPRC"] = ini
        argv = [php, '-q']
        
        if not "PHPRC" in env:
            # php will look in cwd for php.ini also.
            cwd = os.path.dirname(php)
        else:
            cwd = None
        # PHP sets stdin, stdout to binary on it's end.  If we do not do
        # the same, some wierd things can happen, most notably with shell
        # comments that end with \r\n.
        try:
            p = process.ProcessOpen(argv, mode='b', env=env, cwd=cwd)
            p.stdin.write(phpCode)
            p.stdin.close()
            output = p.stdout.read()
            # For some reason, PHP linter causes an exception on close
            # with errno = 0, will investigate in PHP later.
            try:
                p.close()
            except IOError, e:
                if e.errno == 0:  
                    pass
                else:
                    raise
        except OSError, e:
            if e.errno == 0 or e.errno == 32:
                # this happens if you are playing
                # in prefs and change the executable, but
                # not the ini file (ie ini is for a different
                # version of PHP)
                log.error("Caught expected PHP execution error, don't worry be happy: %s", e.strerror)
                return ""
            else:
                log.error("Caught PHP execution exception: %s", e.strerror)
                return ""
        return output.strip()

    def _parsedOutput(self, out):
        """Parse the given output from running PHP.

        If it looks like there is no relevant output, the empty string
        is returned.

        XXX This makes the assumption that all interesting output is on
            one line because only the last non-empty line is used. Any
            leading lines are presumed to be load time errors from PHP.
        """
        if not out: return ""
        # If PHP has load time errors, such as failure loading extension
        # dll's, it spits out a bunch of errors first, then the last
        # line has what we're asking for.  So only get the last line of
        # output.
        lines = re.split('\r\n|\n|\r',out) #XXX should use .splitlines() here
        #XXX Shane, you are doing exactly what you think you are here.
        #    If "out" is a bunch of error lines followed by a blank line
        #    then the last error line is returned here.
        # depending on version of php, we may have a
        # blank last line, check for it.
        if not lines[-1]:
            del lines[-1]
            if not lines: return ""
        return lines[-1]
        
    def _GetPHPConfigVar(self, varName):
        # always output a newline, some versions of php need it
        out = self._GetPHPOutput("<?php echo(get_cfg_var('%s').\"\\n\"); ?>"\
                                  % varName)
        return self._parsedOutput(out)
    
    def _GetPHPIniVar(self, varName):
        # always output a newline, some versions of php need it
        out = self._GetPHPOutput("<?php echo(ini_get('%s').\"\\n\"); ?>"\
                                  % varName)
        return self._parsedOutput(out)

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        phpExes = self.FindInstallationExecutables()
        phpInstallationPaths = [self.getInstallationPathFromBinary(p)\
                                   for p in phpExes]            
        return phpInstallationPaths

    def FindInstallationExecutables(self):
        if sys.platform.startswith('win'):
            exts = ['.exe']
        else:
            exts = None
        phpExes = which.whichall('php', exts=exts, path=self._userPath) + \
               which.whichall('php-cgi', exts=exts, path=self._userPath) + \
               which.whichall('php4', exts=exts, path=self._userPath) + \
               which.whichall('php-cli', exts=exts, path=self._userPath)
        return phpExes

    def getInstallationPathFromBinary(self, binaryPath):
        return os.path.dirname(binaryPath)

    def get_executablePath(self):
        return self._GetPHPExeName()

    def set_executablePath(self, exe):
        self.installationPath = None
        self._configPath = None
        self._executable = exe
        
    def get_haveLicense(self):
        return 1

    def get_version(self):
        # always output a newline, some versions of php need it
        out = self._GetPHPOutput("<?php echo(phpversion().\"\\n\"); ?>")
        return self._parsedOutput(out)

    def get_valid_version(self):
        version = self.get_version()
        if version:
            try:
                version = version.split(".")

                if string.atoi(version[0]) > 4:
                    return 1
                # else must be ge 4.0.5
                elif string.atoi(version[0]) < 4:
                    return 0
                elif string.atoi(version[1]) > 0:
                    return 1
                elif string.atoi( re.match(r"(\d+)", version[2]).group() ) > 4:
                    return 1
            except ValueError,e:
                pass
        return 0
        
    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        Nada for PHP. There is no *standard* local documentation link or any
        real de facto standard.
        """
        return None

    def get_webHelpURL(self):
        return "http://aspn.ActiveState.com/ASPN/PHP/Reference"

    # additional koIPHPInfoEx routines
    # XXX php takes a directory as an argument to define where to find
    # the ini file, but if you query php for this, it returns a file
    def get_cfg_file_path(self):
        out = self._GetPHPConfigVar("cfg_file_path")
        return self._parsedOutput(out)
    
    def set_cfg_file_path(self,path):
        self._configPath = path
        
    def get_include_path(self):
        out = self._GetPHPIniVar("include_path")
        return self._parsedOutput(out)
    
    def set_installationPath(self,value):
        if not os.path.isdir(value):
            self.installationPath = os.path.dirname(value)
        else:
            self.installationPath = value
        
    def GetIncludePathArray(self):
        includePath = self.get_include_path().split(os.pathsep)
        # cull out any empty entries (resulting from, say, include_path="a;;b")
        includePath = [path for path in includePath if path]
        return includePath
    
    def get_extension_dir(self):
        out = self._GetPHPIniVar("extension_dir")
        return self._parsedOutput(out)

    def get_isDebuggerExtensionLoadable(self):
        # always output a newline, some versions of php need it
        out = self._GetPHPOutput("<?php echo extension_loaded('Active Debugger')?\"Yes\\n\":\"No\\n\"; ?>")
        out = self._parsedOutput(out)
        return out == "Yes"
    
class KoPHPInfoEx(KoPHPInfoInstance):
    _com_interfaces_ = [components.interfaces.koIPHPInfoEx]
    _reg_clsid_ = "ea1519a8-4e4d-4767-aec4-2f0342c33e7a"
    _reg_contractid_ = "@activestate.com/koPHPInfoEx;1"
    _reg_desc_ = "PHP Information"

    def _GetPHPExeName(self):
        if self.prefService.prefs.hasStringPref("phpDefaultInterpreter") and\
           self.prefService.prefs.getStringPref("phpDefaultInterpreter"):
            return self.prefService.prefs.getStringPref("phpDefaultInterpreter")
        return self._findPHP()

    def _getInterpreterConfig(self):
        if self.prefService.prefs.hasStringPref("phpConfigFile") and\
           self.prefService.prefs.getStringPref("phpConfigFile"):
            return self.prefService.prefs.getStringPref("phpConfigFile")
        return None

class KoCVSInfoEx(KoAppInfoEx):
    _com_interfaces_ = [components.interfaces.koIAppInfoEx]
    _reg_clsid_ = "C3A7A887-D0D3-426A-8C67-2CC3E2946636"
    _reg_contractid_ = "@activestate.com/koCVSInfoEx;1"
    _reg_desc_ = "CVS Information"

    def __init__(self):
        KoAppInfoEx.__init__(self)
        self._userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)

    def _getCVSExeName(self):
        if not self.installationPath:
            cvsExe = self.prefService.prefs.getStringPref("cvsExecutable")
            if cvsExe:
                self.installationPath = self.getInstallationPathFromBinary(cvsExe)
                return os.path.basename(cvsExe)
            else:
                paths = self.FindInstallationPaths()
                if len(paths) > 0:
                    self.installationPath = paths[0]
                else:
                    return None
        
        if sys.platform.startswith('win'):
            return 'cvs.exe'
        else:
            return 'cvs'

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        if sys.platform.startswith('win'):
            exts = ['.exe']
        else:
            exts = None
        cvss = which.whichall("cvs", exts=exts, path=self._userPath)
        cvsInstallationPaths = [self.getInstallationPathFromBinary(cvs)\
                                for cvs in cvss]
        return cvsInstallationPaths

    def getInstallationPathFromBinary(self, binaryPath):
        if sys.platform.startswith("win"):
            return os.path.dirname(binaryPath)
        else:
            return os.path.dirname(os.path.dirname(binaryPath))

    def get_executablePath(self):
        if not self.executablePath:
            exename = self._getCVSExeName()
            if not exename: return None
            if sys.platform.startswith("win"):
                self.executablePath = os.path.join(self.installationPath, exename)
            else:
                self.executablePath = os.path.join(self.installationPath, "bin", exename)
        return self.executablePath

    def set_executablePath(self,path):
        self.executablePath = path
        if sys.platform.startswith('win'):
            self.installationPath = os.path.dirname(path)
        else:
            self.installationPath = os.path.dirname(os.path.dirname(path))

    def get_haveLicense(self):
        return 1

    def get_version(self):
        """A CVS version include not only the standard 1.2.3-type numbers
        but also the "build family", of which CVSNT is a different one.
        For example:
            1.11.2 CVS
            1.11.1.3 CVSNT
        Returns None if the version cannot be determined.
        """
        cvsExe = self.get_executablePath()
        if not cvsExe: return None
        p = process.ProcessOpen([cvsExe, '-v'])
        p.stdin.close()
        output = p.stdout.read()
        error = p.stderr.read()
        retval = p.wait()
        p.close()
        
        versionRe = re.compile(r'Concurrent Versions System '\
                                '\((?P<family>.+?)\) '\
                                '(?P<version>[\d\.\w]+?) ', re.MULTILINE)
        match = versionRe.search(output)
        if match:
            version = "%s %s" % (match.group('version'),
                                 match.group('family'))
            return version
        else:
            log.warn('Could not determine CVS version [%s] "%s"', cvsExe, output)
            return None
 
    def get_buildNumber(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def get_localHelpFile(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def get_webHelpURL(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)



#---- self test code

if __name__ == "__main__":
    def getCOMAttribute(obj, property, default = 'not implemented'):
        try:
            return getattr(obj, property)
        except COMException, e:
            if e.errno == nsError.NS_ERROR_NOT_IMPLEMENTED:
                return default
            else:
                raise
                    
    for app in ["Perl", "Python", "PHP", "TclPro"]:
        appInfoExe = components.classes["@activestate.com/ko%sInfoEx;1"%app]\
                   .createInstance()
        
        installations = appInfoExe.FindInstallationPaths()
        for installation in installations:
            appInfoExe.installationPath = installation
            print "+------ %s installation: %s" % (app, appInfoExe.installationPath)
            print "| haveLicense: %s" % getCOMAttribute(appInfoExe, 'haveLicense')
            print "| executable location: %s" % getCOMAttribute(appInfoExe, 'executablePath')
            print "| version: %s" % getCOMAttribute(appInfoExe, 'version')
            print "| localHelpFile: %s" % getCOMAttribute(appInfoExe, 'localHelpFile')
            print "| webHelpURL: %s" % getCOMAttribute(appInfoExe, 'webHelpURL')
            
            if app == "TclPro":
                print "|\t+------ TclPro extra features ------"
                print "|\t| tclsh_path: %s" % getCOMAttribute(appInfoExe, 'tclsh_path')
                print "|\t| wish_path: %s" % getCOMAttribute(appInfoExe, 'wish_path')
                print "|\t+---------------------------------"
            elif app == "PHP":
                print "|\t+------ PHP extra features ------"
                print "|\t| cfg_file_path: %s" % appInfoExe.cfg_file_path
                print "|\t| include_path: %s %s" % (appInfoExe.include_path,
                                                    appInfoExe.GetIncludePathArray())
                print "|\t| extension_dir: %s" % appInfoExe.extension_dir
                print "|\t+---------------------------------"
            
            print "+---------------------------------"
    


