#!python
# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

# The implementation of the Komodo Find and Replace service

import sys
import string
import re
import logging

from xpcom import components, nsError, ServerException, COMException
from xpcom.server import UnwrapObject


#---- globals

log = logging.getLogger("find")
#log.setLevel(logging.DEBUG)

# find types
FOT_SIMPLE = components.interfaces.koIFindOptions.FOT_SIMPLE
FOT_WILDCARD = components.interfaces.koIFindOptions.FOT_WILDCARD
FOT_REGEX_PYTHON = components.interfaces.koIFindOptions.FOT_REGEX_PYTHON

# case sensitivity options
FOC_INSENSITIVE = components.interfaces.koIFindOptions.FOC_INSENSITIVE
FOC_SENSITIVE = components.interfaces.koIFindOptions.FOC_SENSITIVE
FOC_SMART = components.interfaces.koIFindOptions.FOC_SMART

# services setup in Find Service ctor
prefSvc = None
lastErrorSvc = None


#---- find and replace components

class KoFindResult:
    _com_interfaces_ = components.interfaces.koIFindResult
    _reg_desc_ = "Find Result"
    _reg_clsid_ = "{0D889F34-0369-4363-BAAB-7465D14EE421}"
    _reg_contractid_ = "@activestate.com/koFindResult;1"

    def __init__(self, url, start, end, value):
        self.url = url
        self.start = start
        self.end = end
        self.value = value


class KoReplaceResult:
    _com_interfaces_ = [components.interfaces.koIFindResult,
                        components.interfaces.koIReplaceResult]
    _reg_desc_ = "Replace Result"
    _reg_clsid_ = "{53056470-F2F0-4a05-B8FF-36DE117C9741}"
    _reg_contractid_ = "@activestate.com/koReplaceResult;1"

    def __init__(self, url, start, end, value, replacement):
        self.url = url
        self.start = start
        self.end = end
        self.value = value
        self.replacement = replacement


class KoFindOptions:
    _com_interfaces_ = components.interfaces.koIFindOptions
    _reg_desc_ = "Find and Replace Session Options"
    _reg_clsid_ = "{3C4AC462-638E-4fa5-A264-3540FEFA558D}"
    _reg_contractid_ = "@activestate.com/koFindOptions;1"

    def __init__(self):
        global prefSvc
        self.patternTypePrefName = "find-patternType"
        self.caseSensitivityPrefName = "find-caseSensitivity"
        self.matchWordPrefName = "find-matchWord"
        self.searchBackwardPrefName = "find-searchBackward"
        self.preferredContextTypePrefName = "find-preferredContextType"

        self.patternType = prefSvc.prefs.getLongPref(self.patternTypePrefName)
        self.caseSensitivity = prefSvc.prefs.getLongPref(self.caseSensitivityPrefName)
        self.matchWord = prefSvc.prefs.getBooleanPref(self.matchWordPrefName)
        self.searchBackward = prefSvc.prefs.getBooleanPref(self.searchBackwardPrefName)
        self.preferredContextType = prefSvc.prefs.getLongPref(self.preferredContextTypePrefName)

    def set_patternType(self, value):
        global prefSvc
        self.patternType = value
        return prefSvc.prefs.setLongPref(self.patternTypePrefName, value)

    def set_caseSensitivity(self, value):
        global prefSvc
        self.caseSensitivity = value
        return prefSvc.prefs.setLongPref(self.caseSensitivityPrefName, value)

    def set_matchWord(self, value):
        global prefSvc
        self.matchWord = value
        return prefSvc.prefs.setBooleanPref(self.matchWordPrefName, value)

    def set_searchBackward(self, value):
        global prefSvc
        self.searchBackward = value
        return prefSvc.prefs.setBooleanPref(self.searchBackwardPrefName, value)

    def set_preferredContextType(self, value):
        global prefSvc
        self.preferredContextType = value
        return prefSvc.prefs.setLongPref(self.preferredContextTypePrefName, value)


class KoFindService:
    _com_interfaces_ = components.interfaces.koIFindService
    _reg_desc_  = "Find and Replace Service"
    _reg_clsid_ = "{3582DE9B-FA36-4787-B3D3-6B9F94EB4AD0}"
    _reg_contractid_ = "@activestate.com/koFindService;1"

    def __init__(self):
        import findlib

        global prefSvc, lastErrorSvc
        prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                  .getService(components.interfaces.koIPrefService)
        lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)

        # Define mappings from koIFinder.idl option enums to findlib option
        # values.
        self.patternTypeMap = {    # koIFinder.patternType -> findlib.patternType
            FOT_SIMPLE:      "simple",
            FOT_WILDCARD:    "wildcard",
            FOT_REGEX_PYTHON:"regex-python",
        }
        self.caseMap = {   # koIFinder.caseSensitivity -> findlib.case
            FOC_INSENSITIVE: "insensitive",
            FOC_SENSITIVE:   "sensitive",
            FOC_SMART:       "smart",
        }

        # load the find and replace options
        self.options = KoFindOptions()
       
    def find(self, url, text, pattern, startOffset):
        """Return the result of searching for the first "pattern" in "text".
        Return value is a KoFindResult or None.
        """
        import findlib
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            result = findlib.find(text, pattern, startOffset=startOffset,
                                  patternType=patternType, case=case,
                                  searchBackward=self.options.searchBackward,
                                  matchWord=self.options.matchWord);
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        if result:
            retval = KoFindResult(url=url, start=result.start, end=result.end,
                                  value=result.value)
        else:
            retval = None
        return retval

    def findall(self, url, text, pattern):
        """Return all the results of searching for "pattern" in "text".
        Return value is a list of KoFindResult's.
        """
        import findlib
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            results = findlib.findall(text, pattern, patternType=patternType,
                                      case=case,
                                      matchWord=self.options.matchWord);
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        retvals = []
        for result in results:
            retval = KoFindResult(url=url, start=result.start, end=result.end,
                                  value=result.value)
            retvals.append(retval)
        return retvals

    def replace(self, url, text, pattern, replacement, startOffset):
        """Return a result indicating how to replace the first "pattern" in
        "text" with "replacement".

        Return value is a KoReplaceResult or None.
        """
        import findlib
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            result = findlib.replace(text, pattern, replacement,
                                     startOffset=startOffset,
                                     patternType=patternType, case=case,
                                     searchBackward=self.options.searchBackward,
                                     matchWord=self.options.matchWord);
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        if result:
            retval = KoReplaceResult(url=url,
                                     start=result.start, end=result.end,
                                     value=result.value,
                                     replacement=result.replacement)
        else:
            retval = None
        return retval

    def replaceall(self, url, text, pattern, replacement):
        """Return a result indicating how to replace the first "pattern" in
        "text" with "replacement".
        
        DEPRECATED: use replaceallex()

        Return value is a list of KoReplaceResult's.
        """
        import findlib
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            results = findlib.replaceall(text, pattern, replacement,
                                         patternType=patternType,
                                         case=case,
                                         searchBackward=self.options.searchBackward,
                                         matchWord=self.options.matchWord);
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        retvals = []
        for result in results:
            retval = KoReplaceResult(url=url,
                                     start=result.start, end=result.end,
                                     value=result.value,
                                     replacement=result.replacement)
            retvals.append(retval)
        return retvals

    def findallex(self, url, text, pattern, resultsView, contextOffset,
                  scimoz):
        """Feed all occurrences of "pattern" in the "text" into the given
        koIFindResultsView.

            "url" is the viewId.
            "resultsView" is null or a koIFindResultView instance on which
                the replace results should be logged via the Add*() methods.
            "contextOffset" is text's offset into the scimoz buffer. This
                is only used if resultsView is specified.
            "scimoz" is the ISciMoz interface for current view. This is only
                used if resultsView is specified.
        
        No return value.
        """
        import findlib
        log.info("findallex(url='%s', text=%r, pattern='%s', resultsView,"
                 "contextOffset=%s, scimoz)", url, text[:20]+text[-20:],
                 pattern, contextOffset)

        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            matches = findlib.findallex(
                text, pattern,
                patternType=patternType, case=case,
                matchWord=self.options.matchWord)
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        if matches:
            resultsView = UnwrapObject(resultsView)
            for match in matches:
                value = match.group()
                startIndex = match.start() + contextOffset
                endIndex = startIndex + len(value)
                startLineNum = scimoz.lineFromPosition(startIndex)
                endLineNum = scimoz.lineFromPosition(endIndex)
                contextStartPos = scimoz.positionFromLine(startLineNum)
                contextEndPos = scimoz.getLineEndPosition(endLineNum)
                context = scimoz.getTextRange(contextStartPos, contextEndPos)
                #context = scimoz.getTextRange(startIndex-10, endIndex+10)
                resultsView.AddFindResult(
                    url, startIndex, endIndex, value,
                    url, # fileName (currently url/viewId is the displayName)
                    startLineNum + 1,
                    startIndex - scimoz.positionFromLine(startLineNum),
                    context)

    def findalllines(self, url, text, pattern, contextOffset, scimoz):
        """Return a lines on which "pattern" is found.

            "url" is the viewId.
            "contextOffset" is text's offset into the scimoz buffer. This
                is only used if resultsView is specified.
            "scimoz" is the ISciMoz interface for current view. This is only
                used if resultsView is specified.
        
        Returns a list of line numbers (0-based).
        """
        import findlib
        
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            matches = findlib.findallex(
                text, pattern,
                patternType=patternType, case=case,
                matchWord=self.options.matchWord)
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        lines = {}  # use dict to avoid dupes
        for match in matches:
            startIndex = match.start() + contextOffset
            startLineNum = scimoz.lineFromPosition(startIndex)
            lines[startLineNum] = 1
        return lines.keys()

    def replaceallex(self, url, text, pattern, replacement, session,
                     resultsView, contextOffset, scimoz):
        """Replace all occurrences of "pattern" in the given text.

            "url" is the viewId.
            "session" is the koIFindSession instance.
            "resultsView" is null or a koIFindResultView instance on which
                the replace results should be logged via the Add*() methods.
            "contextOffset" is text's offset into the scimoz buffer. This
                is only used if resultsView is specified.
            "scimoz" is the ISciMoz interface for current view. This is only
                used if resultsView is specified.
        
        Returns:
            1. the replacement text if there was a change, otherwise null
            2. the number of replacements made
        """
        import findlib
        
        # Determine a "skip-zone" in the current URL. I.e., the area already
        # covered by the user in this find session.
        session = UnwrapObject(session)
        secondLastHit = session.GetSecondLastFindResult()
        if not secondLastHit or secondLastHit.url != url:
            skipZone = None
        elif session.wrapped:
            # Skip zone is from 0 to end of second last hit and from
            # fileStartPos to end of file.
            if self.options.searchBackward:
                skipZone = [(0, session.fileStartPos),
                            (secondLastHit.start, len(text))]
            else:
                skipZone = [(0, secondLastHit.end),
                            (session.fileStartPos, len(text))]
        else:
            # Skip zone is from the fileStartPos to the end of the second
            # last hit. (_Second_ last, because the current hit (the "last")
            # one is still a candidate for replacement.)
            if self.options.searchBackward:
                skipZone = [(secondLastHit.start, session.fileStartPos)]
            else:
                skipZone = [(session.fileStartPos, secondLastHit.end)]
        
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            replacementText, numReplacements, matches = findlib.replaceallex(
                text, pattern, replacement,
                patternType=patternType, case=case,
                searchBackward=self.options.searchBackward,
                matchWord=self.options.matchWord, skipZone=skipZone,
                wantMatches=resultsView is not None)
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        if matches:
            resultsView = UnwrapObject(resultsView)
            for match in matches:
                repl = match.expand(replacement)
                startIndex = match.start() + contextOffset
                endIndex = startIndex + len(repl)
                startLineNum = scimoz.lineFromPosition(startIndex)
                endLineNum = scimoz.lineFromPosition(endIndex)
                contextStartPos = scimoz.positionFromLine(startLineNum)
                contextEndPos = scimoz.getLineEndPosition(endLineNum)
                context = scimoz.getTextRange(contextStartPos, contextEndPos)
                #context = scimoz.getTextRange(startIndex-10, endIndex+10)
                resultsView.AddReplaceResult(
                    url, startIndex, endIndex,
                    match.group(), # value
                    repl, # replacement string
                    url, # fileName (currently url/viewId is the displayName)
                    startLineNum + 1,
                    startIndex - scimoz.positionFromLine(startLineNum),
                    context)

        return replacementText, numReplacements


