"""Test Komodo's koILicense component."""

import os
import sys
import unittest
from xpcom import components

#---- globals

licSvc = components.classes["@activestate.com/koLicense;8"]\
         .getService(components.interfaces.koILicense)


#---- support routines

def _getLicenseFileName():
    if sys.platform.startswith("win"):
        from win32com.shell import shellcon, shell
        base = shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0)
        licDir = os.path.join(base, "ActiveState")
    else:
        licDir = os.path.expanduser(os.path.join("~", ".ActiveState"))
    return os.path.join(licDir, "ActiveState.lic")


#---- test cases

class _BaseLicenseTestCase(unittest.TestCase):
    def setUp(self):
        # Install the test ActiveState.lic
        self._licName = _getLicenseFileName()
        # Move the existing license file out of the way.
        self._tmpLicName = self._licName + ".testtmp"
        if os.path.exists(self._licName):
            os.rename(self._licName, self._tmpLicName)

        if self.license is None:
            pass # no license
        else:
            fout = open(self._licName, 'w')
            fout.write(self.license)
            fout.close()
    
    def tearDown(self):
        # Restore the old license.
        if os.path.exists(self._licName):
            os.unlink(self._licName)
        if os.path.exists(self._tmpLicName):
            os.rename(self._tmpLicName, self._licName)

    def test_checkex(self):
        licSvc = components.classes["@activestate.com/koLicense;8"]\
                 .getService(components.interfaces.koILicense)
        featureNames = ["Komodo-1.2", "Komodo", "ASPN Komodo"]
        rv, featureName, feature = licSvc.checkEx(featureNames,
                                                  2001, 9, 15)
        self.failUnless(rv == self.expectedRetval,
                        "%r != %r" % (rv, self.expectedRetval))
        for name, expected in self.expectedValues.items():
            actual = licSvc.fetchValue(feature, name)
            self.failUnless(actual == expected)
        if self.expectedExpiration:
            expirationDate = licSvc.daysUntil(feature, "ExpirationDate")
            self.failUnless(expirationDate == self.expectedExpiration,
                "%r != %r" % (expirationDate, self.expectedExpiration))


class ValidLicenseTestCase(_BaseLicenseTestCase):
    license = """\
c107769db37d02e5dc390aec3b6a38b7|Komodo|ExpirationDate#99-99-9999|IssuedDate#29-01-2002|LicenseType#Commercial|Licensee#Jennifer Zickerman|NumberOfCPUs#1|SerialNo#S6BACB7789|Token#55072098a5904a879d9821795c364cbd|UpdatesUntil#29-01-2003|UserEmail#jenniferz@ActiveState.com|UserId#96206|UserName#Jennifer Zickerman
"""
    expectedRetval = licSvc.LIC_VALID
    expectedValues = {
        'LicenseType': "Commercial",
        'Licensee': "Jennifer Zickerman",
    }
    expectedExpiration = 9999


class NoLicenseTestCase(_BaseLicenseTestCase):
    license = None
    expectedRetval = licSvc.LIC_NOFILE
    expectedValues = {}
    expectedExpiration = None


class NoFeatureLicenseTestCase(_BaseLicenseTestCase):
    license = """\
151c4c016ebd36c84887be7ab08e53be|TclPro|ExpirationDate#99-99-9999|IssuedDate#11-10-2001|LicenseType#Commercial|Licensee#Lena G|NumberOfCPUs#1|SerialNo#SD55296F66|Token#R0legeqetSnx3LTUCOUSeA|UserEmail#LenaG@ActiveState.com|UserId#96243|UserName#Lena G
"""
    expectedRetval = licSvc.LIC_NOFEATURE
    expectedValues = {}
    expectedExpiration = None


class ExpiredLicenseTestCase(_BaseLicenseTestCase):
    license = """\
b13637eef781c42526e2f97b7049d559|ASPN Komodo|LicenseType#Commercial|IssuedDate#15-06-2001|Licensee#Komodo Developer|UserName#Komodo Developer|Token#j9QAbJtK49OVJzAOy1OH2Q|ExpirationDate#14-06-2001|NumberOfCPUs#1|UserEmail#dev-komodo@ActiveState.com|SerialNo#SC239A35D1|UserId#114217
dbeca8cb66f06fd7849101b3f4108bcf|Komodo|LicenseType#Commercial|IssuedDate#15-06-2001|Licensee#Komodo Developer|UserName#Komodo Developer|Token#j9QAbJtK49OVJzAOy1OH2Q|ExpirationDate#14-06-2001|NumberOfCPUs#1|UserEmail#dev-komodo@ActiveState.com|SerialNo#SC239A35D1|UserId#114217
7bc261c8bd4968be249d8fb10ed15aac|Subscription|LicenseType#Commercial|IssuedDate#15-06-2001|Licensee#Komodo Developer|UserName#Komodo Developer|Token#j9QAbJtK49OVJzAOy1OH2Q|ExpirationDate#14-06-2001|NumberOfCPUs#1|UserEmail#dev-komodo@ActiveState.com|SerialNo#SC239A35D1|UserId#114217
"""
    expectedRetval = licSvc.LIC_EXPIRED
    expectedValues = {}
    expectedExpiration = None


#---- mainline

def suite():
    suites = []
    suites.append( unittest.makeSuite(ValidLicenseTestCase) )
    suites.append( unittest.makeSuite(NoLicenseTestCase) )
    suites.append( unittest.makeSuite(NoFeatureLicenseTestCase) )
    suites.append( unittest.makeSuite(ExpiredLicenseTestCase) )
    return unittest.TestSuite(suites)

def test_main():
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(suite())

if __name__ == "__main__":
    test_main()


